<?php

namespace Softneta\MedDream\Core;


/** @brief Validates actions and UIDs passed to various endpoints.

	Search endpoints like <tt>Routes.php?cmd=loadStudySearch</tt> might receive an
	action to filter specific Patient ID values. Valid use cases are a filtered
	strict-matching search for HIS integration by patient, and a similar search
	for patient history. All remaining situations, including unexpected absence
	of the action, are considered hacking attempts.

	With @c $pacs='PacsOne', one particular hacking attempt (manual opening
	of search.html results in blind passing of any action) is partially covered
	by implementation of a %Study UID filter in @c pacs/PacsImplPacsone/Search.php.
	But, there are no Series UID nor SOP Instance UID filters for corresponding
	actions, and the endpoint might be called directly. Other PACSes face even
	greater security risks.
 */
class ActionFilter
{
	/** @brief Possibility to disable the series/image UID filter.

		If @c false, validateAtSeriesLevel() and validateAtImageLevel() will immediately
		succeed regardless of parameters. This is needed for Export function with PGW
		plugins, as neither Routes.php?cmd=loadStudy nor Routes.php?cmd=getStudyStructure
		is called in that case.
	 */
	const SERIES_IMAGES_FILTER_ENABLED_DFL = true;

	/** @brief Is the Patient History functionality enabled? */
	const PATIENT_HISTORY_ENABLED_DFL = true;

	/** @brief See STRICTLY_MATCH_PATIENT_FROM_ACTION in pacs/PacsImplPacsone/Search.php.

		Under PacsOne there is a choice whether integration by Patient ID also lists "similar"
		IDs (usually due to coerced attributes). If it does, then validateAtStudyLevel() might
		see a mismatch against the initial ID; change to `false` in that case.

		Currently we're reluctant to duplicate database behavior of that non-strict matching.
		Caching UIDs from search results might be a better idea but is complicated by use of
		the same search endpoint for the patient history.
	 */
	const STRICTLY_MATCH_STUDY_PATIENT_FROM_ACTION_DFL = true;

	protected $log;                                 /**< @brief An instance of Logging. */
	protected $areUidFiltersEnabled;                /**< @brief See SERIES_IMAGES_FILTER_ENABLED_DFL */
	protected $isPatientHistoryEnabled;             /**< @brief See PATIENT_HISTORY_ENABLED_DFL */
	protected $strictlyMatchStudyPatientFromAction; /**< @brief See STRICTLY_MATCH_STUDY_PATIENT_FROM_ACTION_DFL */


	/** @brief Constructor.

		@param Logging               $log  Instance of Logging
	 */
	public function __construct(Logging $log)
	{
		$this->log = $log;

		$this->isPatientHistoryEnabled = self::PATIENT_HISTORY_ENABLED_DFL;
		$this->areUidFiltersEnabled = self::SERIES_IMAGES_FILTER_ENABLED_DFL;
		$this->strictlyMatchStudyPatientFromAction = self::STRICTLY_MATCH_STUDY_PATIENT_FROM_ACTION_DFL;
	}


	/** @brief Validate a search endpoint action against a cached one.

		@param array $passedAction  Action to validate (might be updated).
		@param array $cachedAction  An initial cached action created by external.php.

		@return string  Error message to be shown to the user (empty if success).

		@note Ensure a readable session for any stored Patient ID constraints.
	 */
	public function check(&$passedAction, $cachedAction)
	{
		if (!isset($cachedAction['option']))	/* likely empty $cachedAction as no HIS integration is in effect */
		{
			return '';
		}
		if (strtolower($cachedAction['option']) == 'patient')
		{
			/* for PATIENT actions, we enforce consistency with external.php
			   (in fact, it would be better to ignore passed values altogether)
			 */
			$pat = isset($passedAction['entry']) ? $passedAction['entry'] : '';
			$pat2 = isset($cachedAction['entry']) ? $cachedAction['entry'] : '';
			if ($pat != $pat2)
			{
				$this->log->asWarn('security override: replacing passed action ' . var_export($passedAction, true) .
					' with cached action ' . var_export($cachedAction, true));
				$passedAction = $cachedAction;
			}
		}
		else
		{
			/* for other ones (STUDY, SERIES, IMAGE), $action *must* exist and must
			   refer to values cached by ?cmd=getStudyStructure (or ?cmd=loadStudy).
			   This is a valid usage of patient history, unless the latter is disabled.
			 */
			if (!$this->isPatientHistoryEnabled)
			{
				return 'Patient history is disabled';
			}
			$passedPatients = array();
			if (isset($passedAction['option']) &&
			    (strtolower($passedAction['option']) == 'patient') &&
			    isset($passedAction['entry']))
			{
				$passedPatients = $passedAction['entry'];
			}
			if (!count($passedPatients))
			{
				$this->log->asErr('security: no mandatory filter in ' . var_export($passedAction, true));
				return 'Unauthorized access to patient(s)';
			}
			sort($passedPatients);
			$accessedPatients = isset($_SESSION['patientIds']) ? $_SESSION['patientIds'] : array();
			sort($accessedPatients);
			$common = array_intersect($passedPatients, $accessedPatients);
			if (count($common) < count($passedPatients))
			{
				$this->log->asErr('security: disallowing access to ' . var_export($passedAction, true) .
					' due to cached IDs ' . var_export($accessedPatients, true));
				return 'Unauthorized access to object(s)';
					/* saying "object" instead of "patient" as $action might also contain option=study etc */
			}
		}
		return '';
	}


	/** @brief Validate the given object against a cached action after studyGetMetadata() completes for it.

		@param string        $type    Usually "study", "series" or "image". __Must be lowercase.__
		@param string        $value   UID of the object. Multiple values delimited by @c ','.
		@param array         $action  The action structure with mandatory fields .option (string) and
		                              .entry (numerically indexed array of strings, each can optionally
		                              contain multiple values delimited by @c ','). @c null yields
		                              true, silently.
		@param array $studyStructure  Its .error will receive an error message when validation fails, and
		                              .patientid is used for post-factum validation against the "option=patient"
		                              action.

		@return bool  true: valid, false: invalid

		@note Missing action structure silently yields @c true. Malformed action structure, or
		      @p studyStructure without .patientid, yields @c false.
	 */
	public function validateAtStudyLevel($type, $value, $action, &$studyStructure)
	{
		if (!self::actionsDoExist($action))
		{
//			$this->log->asInfo('security: no actions');
			return true;
		}
		if (is_null($studyStructure))
		{
			$studyStructure = array();
		}
		$result = false;
		if (!isset($action['option']) || !isset($action['entry']) || !count($action['entry']))
		{
			$this->log->asErr("security: can't validate against a malformed cached action: " . var_export($action, true));
		}
		else
		{
			if (!strcasecmp($action['option'], 'patient'))
			{
				/* this obviously means studyGetMetadata() has already been called. In contrast, validation against
				   study/series/image actions is also possible *before* studyGetMetadata().
				 */
				if (!array_key_exists('patientid', $studyStructure))
				{
					if (array_key_exists('error', $studyStructure) && strlen($studyStructure['error']))
					{
						/* .error is present and not empty: nothing to validate, the error message shall propagate
						   without being overwritten (the caller will likely return only the message without any
						   other -- possibly sensitive -- information in the structure)
						 */
						$result = true;
					}
					else
					{
						$this->log->asErr("security: study missing Patient ID, can't validate against " . var_export($action, true));
					}
				}
				else
					if (strcmp($studyStructure['patientid'], $action['entry'][0]))
					{
						if ($this->strictlyMatchStudyPatientFromAction)
						{
							$this->log->asErr("security: blocked attempt to return study with Patient ID '" .
								$studyStructure['patientid'] . "' when initially opened for action " . var_export($action, true));
						}
						else
						{
							$this->log->asWarn("security: returning study with Patient ID '" .
								$studyStructure['patientid'] . "' when initially opened for action " . var_export($action, true));
							$result = true;
						}
					}
					else
					{
						$result = true;
					}
			}
			else
			{
				if (strcasecmp($action['option'], $type))
				{
					$this->log->asErr("security: disallowing access to $type '$value' due to cached action " . var_export($action, true));
					$studyStructure['error'] = 'Unauthorized access to object(s)';
				}
				else
				{
					$valuesPassed = explode(',', $value);
					sort($valuesPassed);
					$valuesCached = array();
					foreach ($action['entry'] as $entry)
					{
						$subarray = explode(',', $entry);
						$valuesCached = array_merge($valuesCached, $subarray);
					}
					sort($valuesCached);
					$common = array_intersect($valuesPassed, $valuesCached);
					if (count($common) < count($valuesPassed))
					{
						$this->log->asErr("security: disallowing access to $type '" . join(';', $valuesPassed) .
							"' due to cached action " . var_export($action, true));
					}
					else
					{
						$result = true;
					}
				}
			}
		}
		if (!$result)
		{
			$studyStructure['error'] = 'Unauthorized access to object(s)';
		}
		return $result;
	}


	/** @brief Validate the given Series UID against a cached list.

		@param string $uid     Series UID to validate
		@param array  $action  The action structure. Empty array or @c null yields success, silently. Otherwise
		                       it must contain fields @c .option and @c .entry.

		@return string  Non-empty when validation fails (a message suitable for the end user)

		@note Ensure a readable session for any stored Series UID constraints.

		Object that failed validation is logged as an error.

		Validation is effectively disabled if there is no action.
	 */
	public function validateAtSeriesLevel($uid, $action)
	{
		if (!$this->areUidFiltersEnabled)
		{
			return '';
		}
		if (!self::actionsDoExist($action))
		{
//			$this->log->asInfo('security: no actions');
			return '';
		}
		if (!isset($action['option']) || !isset($action['entry']))
		{
			$this->log->asErr("security: can't validate against a malformed cached action: " . var_export($action, true));
		}
		else
		{
			$parts = explode('*', $uid);
			$firstComponent = $parts[0];
			if (isset($_SESSION['seriesUids']) && in_array($firstComponent, $_SESSION['seriesUids']))
			{
				return '';
			}
			else
			{
				$this->log->asErr("security: disallowing access to series '$firstComponent'");
			}
		}
		return 'Unauthorized access to object(s)';
	}


	/** @brief Validate the given SOP Instance UID against a cached list.

		@param string $uid    SOP Instance UID to validate
		@param array  $action  The action structure. Empty array or @c null yields success, silently. Otherwise
		                       it must contain fields @c .option and @c .entry.

		@return string  Non-empty when validation fails (a message suitable for the end user)

		@note Ensure a readable session for any stored SOP Instance UID constraints.

		Object that failed validation is logged as an error.

		Validation is effectively disabled if there is no action.
	 */
	public function validateAtImageLevel($uid, $action)
	{
		if (!$this->areUidFiltersEnabled)
		{
			return '';
		}
		if (!self::actionsDoExist($action))
		{
//			$this->log->asInfo('security: no actions');
			return '';
		}
		if (!isset($action['option']) || !isset($action['entry']))
		{
			$this->log->asErr("security: can't validate against a malformed cached action: " . var_export($action, true));
		}
		else
		{
			$parts = explode('*', $uid);
			$firstComponent = $parts[0];
			if (isset($_SESSION['imageUids']) && in_array($firstComponent, $_SESSION['imageUids']))
			{
				return '';
			}
			else
			{
				$this->log->asErr("security: disallowing access to image '$firstComponent'");
			}
		}
		return 'Unauthorized access to object(s)';
	}


	public static function actionsDoExist($actions)
	{
		return !is_null($actions) && count($actions);
	}


	/** @brief Call externalActions() from external.php.

		@param array $result  Receives the "actions" array. It's initialized to an empty array
		                      just before the call, or to @c null before returning @c false.

		@return bool Indicates if external.php is present and a call has been made.
	 */
	public static function getActionsInto(&$result)
	{
		$file = __DIR__ . '/external.php';
		clearstatcache(false, $file);
		if (!@file_exists($file))
		{
			$result = null;
			return false;
		}

		$result = array();
		require_once($file);
		externalActions($result);
		return true;
	}


	/** @brief Cache Patient ID as a constraint related to cached actions of STUDY type.

		@param string $pid  Value to cache

		@note Ensure an R/W session in advance as values will be cached there.
	 */
	public static function markValidPatientId($pid)
	{
		if (!isset($_SESSION['patientIds']))
		{
			$_SESSION['patientIds'] = array();
		}
		if (!in_array($pid, $_SESSION['patientIds']))
		{
			$_SESSION['patientIds'][] = $pid;
		}
	}


	/** @brief Cache a list of Series UIDs as a constraint for SERIES-level endpoints.

		@param array $uidList  Values to cache

		@note Ensure an R/W session in advance as values will be cached there.

		Values will be appended. The only way to clear the list is forgetValidIds() that is called
		automatically during logoff or when HIS integration opens a different object.
	 */
	public static function markValidSeriesUids($uidList)
	{
		if (!isset($_SESSION['seriesUids']))
		{
			$_SESSION['seriesUids'] = array();
		}
		$_SESSION['seriesUids'] = array_unique(array_merge($_SESSION['seriesUids'], $uidList));
	}


	/** @brief Cache a list of SOP Instance UIDs as a constraint for IMAGE-level endpoints.

		@param array $uidList  Values to cache

		@note Ensure an R/W session in advance as values will be cached there.

		Values will be appended. The only way to clear the list is forgetValidIds() that is called
		automatically during logoff or when HIS integration opens a different object.
	 */
	public static function markValidImageUids($uidList)
	{
		if (!isset($_SESSION['imageUids']))
		{
			$_SESSION['imageUids'] = array();
		}
		$_SESSION['imageUids'] = array_unique(array_merge($_SESSION['imageUids'], $uidList));
	}


	/** @brief Remove all cached constraints.

		@note Ensure an R/W session in advance as values will be cached there.
	 */
	public static function forgetValidIds()
	{
		$_SESSION['patientIds'] = array();
		$_SESSION['seriesUids'] = array();
		$_SESSION['imageUids'] = array();
	}
}
