<?php

namespace Softneta\MedDream\Core;


/** @brief Provides data for the About dialog in 6.0+. */
class ProductInfo
{
	protected $log;                 /**< @brief Instance of Logging. */
	protected $tr;                  /**< @brief Instance of Translation. */
	protected $brnd;                /**< @brief Instance of Branding. */
	protected $uc;                  /**< @brief Instance of UpdateChecker. */
	protected $sys;                 /**< @brief Instance of System. */
	protected $backend;             /**< @brief Instance of Backend. The 'Auth' part must be available. */
	protected $usedAsModule;        /**< @brief "isUsedAsModule" from license file. */
	protected $isRebranded;         /**< @brief "isRebranded" from license file. */


	/** @brief Constructor */
	public function __construct(Backend $backend, System $sys, UpdateChecker $uc, Branding $brnd,
		Translation $tr, Logging $log)
	{
		$this->backend = $backend;
		$this->sys = $sys;
		$this->uc = $uc;
		$this->brnd = $brnd;
		$this->tr = $tr;
		$this->log = $log;
	}


	/** @brief Extract human-readable fields from the license file.

		@note Updates @link $usedAsModule @endlink, @link $isRebranded @endlink.
	 */
	protected function fromLicense()
	{
		$xmlStr = $this->sys->license('');
		$xmlStr = str_replace(array("\n", "\r", "\t"), '', $xmlStr);
		$xmlStr = trim(str_replace('"', "'", $xmlStr));
		$xmlObj = @simplexml_load_string($xmlStr);
		$lic = array();
		$lic['connections'] = function_exists('meddream_connections') ? meddream_connections() : false;
		if (isset($xmlObj->maxUsers))
		{
			$lic['maxUsers'] = (string) $xmlObj->maxUsers;
		}
		else
		{
			$lic['maxUsers'] = '';
		}
		if (isset($xmlObj->module))
		{
			$lic['modules'] = explode(',', (string) $xmlObj->module);
		}
		else
		{
			$lic['modules'] = array();
		}
		if (isset($xmlObj->registeredTo))
		{
			$lic['registeredTo'] = (string) $xmlObj->registeredTo;
		}
		else
		{
			$lic['registeredTo'] = '';
		}
		if (isset($xmlObj->updatesTo))
		{
			$lic['updatesTo'] = (string) $xmlObj->updatesTo;
		}
		else
		{
			$lic['updatesTo'] = '';
		}
		if (isset($xmlObj->validTo))
		{
			$lic['validTo'] = (string) $xmlObj->validTo;
		}
		else
		{
			$lic['validTo'] = '';
		}

		if (isset($xmlObj->isRebranded))
		{
			$this->isRebranded = '1' == (string) $xmlObj->isRebranded;
		}
		else
		{
			$this->isRebranded = false;
		}
		if (isset($xmlObj->isUsedAsModule))
		{
			$this->usedAsModule = '1' == (string) $xmlObj->isUsedAsModule;
		}
		else
		{
			$this->usedAsModule = false;
		}

		return $lic;
	}


	/** @brief Detect whether the "VET" module is present.

		@param array $licData  Return value of fromLicense()

		@return bool
	 */
	protected function hasVetModule($licData)
	{
		foreach ($licData['modules'] as $mod)
		{
			$module = strtoupper(trim($mod, "+ \r\n"));
			if ($module == 'VET')
			{
				return true;
			}
		}

		return false;
	}


	/** @brief Indicates if updating to a newer version is possible.

		@param string          $currentVersion    @c $VERSION from sharedData.php
		@param string          $currentUpdatesTo  @c updatesTo from license file

		@return array
	 */
	protected function getUpdateStatus($currentVersion, $currentUpdatesTo)
	{
		/* some safe defaults */
		$result = array(
			'availableNewVersion' => false,
			'requiresNewLicense' => true,
			'isAdmin' => false
		);

		$status = $this->uc->checkCached($currentVersion, $currentUpdatesTo);
		if ($status === false)
		{
			return $result;
		}
		$result['availableNewVersion'] = $status['newerVersionIsAvailable'];
		$result['requiresNewLicense'] = $status['requiresNewLicense'];

		$result['isAdmin'] = $this->backend->pacsAuth->hasPrivilege('root') == 1;

		return $result;
	}


	/** @brief Provide FDA certification-related text. */
	protected function fromFdaCert()
	{
		return $this->tr->translate('AboutForm\lblFdatext', 'For diagnostic use K162011 (device class: 2)');
	}


	/** @brief Extract branding-related data.

		@param array $data  Will change rebrandable elements of this array

		@return bool  Indicates if branding was applied or not
	 */
	protected function applyBranding(&$data)
	{
		if (!$this->isRebranded || !$this->brnd->isValid() || !$this->brnd->active())
		{
			return false;
		}

		$data['rebrandingType'] = $this->usedAsModule ? 'full' : 'partly';

		$data['name'] = $this->brnd->getAttribute('productName');
		$data['version'] = $this->brnd->getAttribute('productVersion');
		$data['contactCompany'] = array(
			'name' => $this->brnd->getAttribute('companyName'),
			'email' => $this->brnd->getAttribute('email'),
			'site' => $this->brnd->getAttribute('site'),
			'address' => array(
				'street' => $this->brnd->getAttribute('streetAddr'),
				'postCode' => $this->brnd->getAttribute('postCode'),
				'city' => $this->brnd->getAttribute('city'),
				'country' => $this->brnd->getAttribute('country')
			)
		);

		return true;
	}


	/** @brief Collect data from license, branding, build constants, etc.

		@return string  Something failed, this is an error message
		@retval array   Collected data
	 */
	public function collectData()
	{
		require __DIR__ . '/sharedData.php';
		$thisVer = '';
		if (isset($VERSION))
		{
			$thisVer = trim($VERSION);
		}

		if (!$this->backend->authDB->isAuthenticated())
		{
			return 'not authenticated';
		}

		$rtn = array();

		/* the license block */
		$lic = $this->fromLicense();
		$rtn['license'] = $lic;

		/* data from the license block is also needed for additional logic related to VET license */
		$isVet = $this->hasVetModule($lic);

		/* the updates block */
		$rtn['updates'] = $this->getUpdateStatus($thisVer, $rtn['license']['updatesTo']);

		/* the product block */
		$product = array(
			'rebrandingType' => 'none',
			'isCertified' => !$this->usedAsModule,
			'name' => $PRODUCT,
			'fdaClarification' => $this->fromFdaCert(),
			'ce' => array(
				'class' => "Directive 93/42/EEC\nClass IIa active medical device",
				'notifiedBodyId' => '2460'
			),
			'version' => $thisVer,
			'releaseDate' => $RELDATE,
			'contactCompany' => array(),
		);
		$rebranded = $this->applyBranding($product);
		if (!$rebranded)
		{
			unset($product['contactCompany']);
		}
		if (!$this->usedAsModule || ($this->usedAsModule && !$rebranded))
		{
			$product['manufacturer'] = array(
				'name' => 'Softneta, UAB',
				'email' => 'info@softneta.com',
				'site' => 'http://www.softneta.com',
				'address' => array(
					'street' => 'K. Baršausko str. 59b',
					'postCode' => 'LT-51423',
					'city' => 'Kaunas',
					'country' => 'Lithuania'
				)
			);
		}
		else
		{
			unset($product['fdaClarification']);
			unset($product['ce']);
		}
		if ($isVet)
		{
			$product['isCertified'] = false;
			unset($product['fdaClarification']);
			unset($product['ce']);
		}
		$rtn['product'] = $product;

		return $rtn;
	}
}
