<?php

namespace Softneta\MedDream\Core;


/** @brief Checks for product updates. */
class UpdateChecker
{
	/** @brief Web address to get the update manifest from. */
	const LATEST_VERSION_WEB_PATH = '/files/meddreamviewer/LATEST';

	/** @brief A cached update manifest newer than this age won't be updated

		@note Ignored when calling checkCached(true) or constructor's @c $cacheAge
		      parameter is set to 0.
	 */
	const DFL_CACHE_AGE = 3600;

	protected $cacheAge;                /**< @brief Expiration age (in seconds) of cache contents */
	protected $downloader;              /**< @brief An instance of BasicRestClient */
	protected $log;                     /**< @brief An instance of Logging */
	protected $constants;               /**< @brief An instance of Constants */

	/** @brief Path to the cache file.

		The file contains a cached response to the remote version query.
		If its age (as per file timestamp) is too small, then a new query
		won't be executed and cacheRemoteVersion() will return immediately.

		The response to the query is reset to @c false on any error,
		including non-200 HTTP response codes. This will be correctly
		preserved and later reacted on.
	 */
	protected $cacheFilePath;


	/** @brief Return the cached remote version.

		@param  string $timestamp  Receives timestamp of the cache file, or @c false
		                           if the file does not exist (nothing is cached)

		@retval false   The query failed, the file is of invalid format or missing
		                (@p $timestamp allows to detect the last situation)
		@retval string  The query was successful and this is the actual server
		                response
	 */
	protected function loadCachedVersion(&$timestamp)
	{
		$rsp = false;

		clearstatcache(false, $this->cacheFilePath);
		$timestamp = @filemtime($this->cacheFilePath);
		if ($timestamp !== false)
		{
			$contents = @file_get_contents($this->cacheFilePath);
			if ($contents === false)
			{
				$this->log->asErr('existing cache file is unreadable: ' . $this->cacheFilePath);
			}
			else
			{
				$encodedFalse = serialize(false);
				if ($contents != $encodedFalse)		/* something else was serialized, let's try to decode */
				{
					$rsp = @unserialize($contents);
					if ($rsp === false)				/* definitely an encoding problem */
					{
						$this->log->asErr('cache file has invalid format: ' . $this->cacheFilePath);
					}
				}
			}
		}

		return $rsp;
	}


	/** @brief Date comparison for checkCached().

		@return boolean  @c true if first date < second date
	 */
	protected function dateBefore($date1, $date2)
	{
		if ($date1 == '0000-00-00')
		{
			return true;
		}

		if ($date2 == '0000-00-00')
		{
			return true;
		}

		$date11 = strtotime($date1);
		$date12 = strtotime($date2);
		return $date11 < $date12;
	}


	/** @brief Initializes various internals.

		@param int             $cacheAge          How many seconds a cached response is reused.
		                                          Negative values (default) are replaced by DFL_CACHE_AGE.
		                                          If you specify zero, then cache is refreshed every time.
		@param string          $cachePath         Override path to the cache file. @c null is replaced
		                                          by a safe default value.
		@param BasicRestClient $client            An object instance of BasicRestClient
		@param Logging         $log               An object instance of Logging
		@param Constants       $constants         An object instance of Constants
	 */
	public function __construct($cacheAge = -1, $cachePath = null,
		BasicRestClient $client = null, Logging $log = null, Constants $constants = null)
	{
		if ($cacheAge < 0)
		{
			$cacheAge = self::DFL_CACHE_AGE;
		}
		$this->cacheAge = $cacheAge;

		if (is_null($cachePath))
		{
			$cachePath = __DIR__ . '/temp/LATEST_VERSION';
		}
		$this->cacheFilePath = $cachePath;

		if (is_null($client))
		{
			$client = new BasicRestClient('GET');
		}
		$this->downloader = $client;

		if (is_null($log))
		{
			$log = new Logging();
		}
		$this->log = $log;

		if (is_null($constants))
		{
			$constants = new Constants();
		}
		$this->constants = $constants;
	}


	/** @brief Query the remote version from server anew.

		@param bool $refresh  @c false: do not perform the query if an already cached value
		                      does not indicate errors and is fresh enough. If @c true, or
		                      @link $cacheAge @endlink is zero, then force the query.
	 */
	public function cacheRemoteVersion($refresh = false)
	{
		$this->log->asDump('begin ' . __METHOD__);

		/* do not proceed if the cached value is "good" but too fresh */
		if (!$refresh && $this->cacheAge)
		{
			$body = $this->loadCachedVersion($ts);
			if ($body !== false)
			{
				$age = time() - $ts;
				if ($age < $this->cacheAge)
				{
					$this->log->asDump("not updating cached value (age $age seconds)");
					return;
				}
			}
		}

		$c = $this->constants;	/* alas we need a local variable to access class constants via an instance */
		$url = $c::HOME_URL . self::LATEST_VERSION_WEB_PATH;
		$this->log->asDump('querying latest MedDream version from ', $url);

		/* download a new license file */
		$body = null;
		$r = $this->downloader->request($url, $body, $details);
		if ($r != 200)
		{
			$this->log->asErr("remote version query: HTTP Status $details, HTTP Body '" .
				addcslashes($body, "\0..\37") . "'");
			$body = false;
		}
		else
		{
			$this->log->asDump("remote version query: HTTP Status $details, HTTP Body ",
				addcslashes($body, "\0..\37"));
		}

		/* cache it in the file */
		$exported = serialize($body);
		if (@file_put_contents($this->cacheFilePath, $exported) === false)
		{
			$this->log->asErr('failed to write to ' . $this->cacheFilePath);
		}
		else
		{
			$this->log->asDump('end ' . __METHOD__);
		}
	}


	/** @brief Decide if this installation is eligible for the update.

		@param string          $currentVersion    @c $VERSION from sharedData.php
		@param string          $currentUpdatesTo  @c updatesTo from license file

		@retval false  There is no cached remote version/date; likely cacheRemoteVersion()
		               never succeeded, or was never called.
		@retval array

		Format of the returned array:

		<ul>
		<li><tt>'newerVersionIsAvailable'</tt> - the remote version is higher (an update is available)
		<li><tt>'requiresNewLicense'</tt> - the remote version requires a license with larger updatesTo
		</ul>
	 */
	public function checkCached($currentVersion, $currentUpdatesTo)
	{
		$this->log->asDump('begin ' . __METHOD__ . '(', $currentVersion, ', ', $currentUpdatesTo, ')');

		$content = $this->loadCachedVersion($ts);
		if ($content === false)
		{
			$this->log->asDump(__METHOD__ . ': cached response is invalid or missing');
			return false;
		}

		/* canonicalize: we must be independent on delimiters */
		if (!strlen($content))
		{
			$this->log->asDump(__METHOD__ . ': cached response is empty');
			return false;		/* no sense in further processing */
		}
		$content = str_replace(array("\r", "\n", "\t", "\v"), ' ', trim($content));
		$lengthNew = -1;
		do
		{
			$lengthOld = $lengthNew;
			$content = str_replace('  ', ' ', $content);
			$lengthNew = strlen($content);
		} while ($lengthNew && ($lengthNew != $lengthOld));
		$this->log->asDump('cleaned up: ', addcslashes($content, "\0..\37"));

		$return = array('newerVersionIsAvailable' => false, 'requiresNewLicense' => true);

		$parts = explode(' ', $content);
		$this->log->asDump('$parts: ', $parts);
		$remoteVersion = '';
		if (isset($parts[0]))
		{
			$remoteVersion = (string) $parts[0];
		}
		$remoteDate = '';
		if (isset($parts[1]))
		{
			$remoteDate = (string) $parts[1];
		}

		if ($remoteVersion == '')
		{
			$this->log->asErr('version number not found in the response');
			if (!@unlink($this->cacheFilePath))		/* invalidate the cache */
			{
				$this->log->asErr('failed to purge the cache: ' . $this->cacheFilePath);
			}
			return $return;
		}

		/* we could do this in the very beginning (and in cacheRemoteVersion(), too) to
		   avoid "needless" activity, however then system testing would be harder.
		   Currently even development builds will periodically download the remote update
		   manifest and keep it in a place easy to check manually.
		 */
		if ($currentVersion == 'DEV')
		{
			$this->log->asInfo('assuming up-to-date for a development version');
			$return['requiresNewLicense'] = false;
			return $return;
		}

		$this->log->asDump('versions: local ', $currentVersion, ', remote ', $remoteVersion);
		$return['newerVersionIsAvailable'] = version_compare($currentVersion, $remoteVersion, '<');

		$this->log->asDump('updatesTo: local ', $currentUpdatesTo, ', remote ', $remoteDate);
		if (!empty($currentUpdatesTo) && !empty($remoteDate))
		{
			/* Both frontends yield an error if updatesTo from license is less than
			   release date hardcoded in their build. We should do the same to minimize
			   misunderstandings.
			 */
			$return['requiresNewLicense'] = $this->dateBefore($currentUpdatesTo, $remoteDate);
		}

		$this->log->asDump('return: ', $return);
		$this->log->asDump('end ' . __METHOD__);
		return $return;
	}
}
